@@ -3,7 +3,6 @@ Python module containing "Master" classes of easy_gui project.
 The classes in here are designed to be subclassed in user applications.
 '''
 import tkinter as tk
-import tkinter.scrolledtext
 from tkinter import ttk
 from tkinter import _tkinter
 from .styles import BaseStyle
@@ -17,6 +16,21 @@ from typing import List, Dict
 
 
 
+def recreate_if_needed(func):
+    '''
+    Decorator used to enable addition of Sections or Widgets after GUI has been created.
+    (that is, can add elements outside of EasyGUI subclass' __init__ method)
+    '''
+    def inner(*args, **kwargs):
+        self = args[0]
+        value = func(*args, **kwargs)
+        if self.root.created:
+            self.root.create()  # need to re-create GUI so that the new elements show up!
+        return value
+    return inner
+
+
+
 class GridMaster():
     def __init__(self):
         self.grid_areas = {}
@@ -59,9 +73,9 @@ class GridMaster():
         if self.grid_areas != {}:
             limits = self.grid_limits()
             for row in range(limits['min_row'], limits['max_row'] + 1):
-                self.grid_rowconfigure(row, weight=1)
+                self.grid_rowconfigure(row, weight=1, minsize=10)
             for col in range(limits['min_col'], limits['max_col'] + 1):
-                self.grid_columnconfigure(col, weight=1)
+                self.grid_columnconfigure(col, weight=1, minsize=10)
 
     def add_grid_row(self, row_name: str):
         if self.grid_configuration == []:
@@ -72,7 +86,7 @@ class GridMaster():
         self.configure_grid(self.grid_configuration)
 
     def grid_limits(self) -> dict:
-        min_row, max_row, min_col, max_col = 500, -500, 500, -500  # arbitrarly large starting points so no risk of surpising row/col not being captured
+        min_row, max_row, min_col, max_col = 500, -500, 500, -500  # arbitrarily large starting points so no risk of surprising row/col not being captured
         for area in self.grid_areas.values():
             if area['first_row'] < min_row:
                 min_row = area['first_row']
@@ -86,11 +100,12 @@ class GridMaster():
 
 
 
-
 class SectionMaster():
     def __init__(self):
         self.sections: dict = {}
+        self.widgets: dict = {}
 
+    @recreate_if_needed
     def add_section(self, name='', title=False, grid_area=None,
                                borderwidth=None, relief=None, tabbed: bool=False, equal_button_width: bool=False, external_section=None):
         '''
@@ -121,6 +136,45 @@ class SectionMaster():
         self.sections[name] = section
         return section
 
+    @recreate_if_needed
+    def add_widget(self, type='label', text='', widget_name=None, grid_area=None, **kwargs):
+        '''
+        Add a Widget object to this Section by calling the add_widget function in widgets.py
+        (Easier to keep the function there as it needs access to all the individual Widget classes.)
+        '''
+        return widgets.add_widget(self, type=type, text=text, widget_name=widget_name, grid_area=grid_area, **kwargs)
+
+    def delete_widget(self, widget_name) -> None:
+        '''
+        Fully delete a widget.
+        Pass without issue if the widget doesn't exist.
+        '''
+        try:
+            self.widgets[widget_name].destroy()
+            del self.widgets[widget_name]
+        except:
+            pass
+
+    def delete_all_widgets(self) -> None:
+        '''
+        Fully delete all child widgets of this section.
+        '''
+        for w_name in list(self.widgets.keys()):
+            self.delete_widget(w_name)
+
+    def _clear_and_recreate_plot(self, mpl_figure, widget_name, grid_area, kwargs):
+        old_widget = self.widgets[widget_name]  # grab reference to widget to be deleted so that its place in dict can be given to new widget
+
+        new_widget = self.add_widget(type='matplotlib', widget_name=widget_name, toolbar=old_widget.toolbar, grid_area=grid_area)
+        new_widget.bindings = old_widget.bindings
+        new_widget.small_figure_warning_given = old_widget.small_figure_warning_given
+        new_widget.position()
+        new_widget.draw_plot(mpl_figure=mpl_figure)
+        new_widget.position()  # have to reposition/create Widget
+
+        old_widget.destroy()  # destroy after new widget is positioned for slightly less flickering
+
+    @recreate_if_needed
     def add_tab(self, name='', **kwargs):
         if not self.tabbed:
             print('Error!  Cannot .add_tab to a Section unless tabbed=True when it is created.')
@@ -163,7 +217,11 @@ class EasyGUI(tk.Tk, GridMaster, SectionMaster):
 
         self.icon(bitmap=os.path.join(os.path.dirname(__file__), 'resources', 'transparent.ico'), default=True)
         self.title('EasyGUI')
+
         self.geometry("300x180+100+60")  # format of "WIDTHxHEIGHT+(-)XPOSITION+(-)YPOSITION"
+        # Instead of setting .geometry, can also set "width", "height" to integer values
+        # and "center" to True in application subclass to size and center window
+
         self.transparent = False
         self.configure(background=self.style.window_color)
 
@@ -173,21 +231,25 @@ class EasyGUI(tk.Tk, GridMaster, SectionMaster):
         # See documention of below WINDOWS options here: https://wiki.tcl-lang.org/page/wm+attributes
         self.wm_attributes('-alpha', alpha)
         self.wm_attributes('-fullscreen', fullscreen)
-        self.wm_attributes('-disabled', disable_interaction)  # disables window interaction for click pass through
-        self.wm_attributes('-toolwindow', toolwindow)  # makes a window with a single close-button (which is smaller than usual) on the right of the title bar
         self.wm_attributes('-topmost', topmost)  # make root window always on top
         self.overrideredirect(overrideredirect)  # hide root window drag bar and close button
+        try:
+            self.wm_attributes('-disabled', disable_interaction)  # disables window interaction for click pass through
+            self.wm_attributes('-toolwindow', toolwindow)  # makes a window with a single close-button (which is smaller than usual) on the right of the title bar
+        except Exception:  # above options only for Windows system
+            pass
 
         s = ttk.Style()
         s.configure('.', background=self.style.widget_bg_color)
         s.configure('.', font=self.style.font)
         s.configure('.', foreground=self.style.text_color)
 
+        self.created = False
 
     def __init_subclass__(cls, **kwargs):
         '''
         Wraps user subclass __init__ to implicitly handle the EasyGUI.__init__ call along with
-        calling .create_gui() after application is fully defined in subclass __init__ method
+        calling .create() after application is fully defined in subclass __init__ method
         '''
         old_init = cls.__init__  # reference to original subclass method so new_init isn't recursive
         def new_init(self, *args, **kwargs):
@@ -198,6 +260,32 @@ class EasyGUI(tk.Tk, GridMaster, SectionMaster):
                 print('\n* Are you passing in kwargs to GUI creation?\n* If so, remember to put a "**kwargs" in the __init__ function!\n')
                 traceback.print_exc()
             self.create()  # populates GUI elements
+
+            # now change window geometry if "width", "height" and/or "center" attributes are set in subclass' __init__ method
+            # seems easier to allow this than forcing self.geometry() usage as that is a bit cryptic and hard to remember
+            # auto-centering by setting self.center = True is also convenient as usually that behavior is desired
+            self.update_idletasks()  # need to run here so any geometry changes from subclass __init__ run before checking sizes
+            current_width, current_height = self.winfo_width(), self.winfo_height()
+            frame_width = self.winfo_rootx() - self.winfo_x()
+            window_width = current_width + 2 * frame_width
+            titlebar_height = self.winfo_rooty() - self.winfo_y()
+            window_height = current_height + titlebar_height + frame_width
+            if hasattr(self, 'width'):
+                window_width = self.width
+            if hasattr(self, 'height'):
+                window_height = self.height
+            if hasattr(self, 'center') and self.center == True:
+                center_x_val = int(self.winfo_screenwidth() / 2 - window_width / 2)
+                center_y_val = int((self.winfo_screenheight() / 2 - window_height / 2))
+                center_y_val -= 30  # offset a little higher than middle since many people have toolbar on bottom of screen
+                if center_x_val < 0: # don't let left side of window go beyond screen if too wide
+                    center_x_val = 0
+                if center_y_val < 0:  # don't let top of window go above screen if too tall
+                    center_y_val = 0
+                self.geometry(f'{window_width}x{window_height}+{center_x_val}+{center_y_val}')
+            elif hasattr(self, 'width') or hasattr(self, 'height'):
+                self.geometry(f'{window_width}x{window_height}')
+
             self.bind_all('<Key>', self.log_keys)
             self.mainloop()  # runs tkinter mainloop
         cls.__init__ = new_init  # overwrite subclass __init__ method
@@ -268,8 +356,12 @@ class EasyGUI(tk.Tk, GridMaster, SectionMaster):
         Positions GUI elements in window.
         May be called recursively by child Sections as elements are positioned.
         '''
-        for name, section in self.sections.items():
-            section.create(force_row=force_row)
+        for child in {**self.widgets, **self.sections}.values():
+            try:
+                child.create(force_row)  # if child is another Section object
+            except AttributeError:
+                child.position(force_row)  # if child is a Widget object
+        self.created = True
 
     def add_menu(self,
                  commands={'File': lambda: print('File button'), 'Edit': lambda: print('Edit button')},
@@ -290,15 +382,89 @@ class EasyGUI(tk.Tk, GridMaster, SectionMaster):
 
         self.config(menu=self.menu)
 
+    def __repr__(self):
+        return 'Main EasyGUI Application'
+
+    def popup(self, *args, **kwargs):
+        '''
+        Returns a context manager for generating a popup window.  Example usage:
+
+        with self.popup() as popup:
+            popup.add_widget('lbl', 'Test1')
+            popup.add_widget('btn', 'Test Button', command_func=lambda *args: print('Test Button clicked'))
+        '''
+        return PopUp(*args, **kwargs)
+
+
+class PopUp(tk.Toplevel, GridMaster, SectionMaster):
+    '''
+    Basically a mini EasyGUI class that inherits from tk.Toplevel instead of tk.Tk.
+    Re-implements basic methods of EasyGUI class so widgets can be added.
+    '''
+    def __init__(self, *args, width: int=300, height: int=180, x: int=120, y: int=80, **kwargs):
+        if kwargs.get('tooltip', False):
+            super().__init__()
+            GridMaster.__init__(self)
+            SectionMaster.__init__(self)
+            self.wm_attributes('-disabled', True)  # disables window interaction for click pass through
+            self.wm_overrideredirect(True)  # removes window
+            self.wm_attributes('-alpha', 0.8)
+            self.geometry(f'{width}x{height}+{x}+{y}')  # format of "WIDTHxHEIGHT+(-)XPOSITION+(-)YPOSITION"
+            self.style = EasyGUI.style
+            self.style.create_font()
+            self.configure(bg=self.style.tooltip_color)
+        else:
+            super().__init__()
+            GridMaster.__init__(self)
+            SectionMaster.__init__(self)
+            self.icon(bitmap=os.path.join(os.path.dirname(__file__), 'resources', 'transparent.ico'), default=True)
+            self.geometry(f'{width}x{height}+{x}+{y}')  # format of "WIDTHxHEIGHT+(-)XPOSITION+(-)YPOSITION"
+            self.style = EasyGUI.style
+            self.style.create_font()
+
+    def __enter__(self):
+        self.created = False
+        return self
+
+    def __exit__(self, *args):
+        self.create()
+
+    @property
+    def root(self):
+        '''Used by downstream elements to reference EasyGUI as root'''
+        return self
+
+    def icon(self, bitmap, default: bool=False) -> None:
+        '''
+        Alternate method to call tk.Tk iconbitmap method using altered path handling
+        so that PyInstaller can package application with specified .ico file.
+        If not default, warning message is printed on failing to locate .ico file.
+        '''
+        try:
+            super().iconbitmap(bitmap=resource_path(bitmap))
+        except _tkinter.TclError:
+            if default:
+                pass  # Pass silently if default .ico not found occurs when using PyInstaller and not adding transparent.ico to "datas"
+            else:
+                print(f'Cannot locate {bitmap}!  If using PyInstaller, be sure to specify this file in "datas".')
+
+    def create(self, force_row=False) -> None:
+        '''Copied from EasyGUI.create'''
+        for name, section in self.sections.items():
+            section.create(force_row=force_row)
+        self.created = True
+
+    @recreate_if_needed
     def add_widget(self, *args, **kwargs):
-        '''
-        Same as Section.add_widget, but works if used on main GUI class without specifying a Section.
-        A Section named "_default" is created if needed and widgets are put within that section.
-        '''
+        '''Copied from EasyGUI.add_widget'''
         if '_default' not in self.sections:
             self.add_section('_default')
         return self.sections['_default'].add_widget(*args, **kwargs)
 
+    def __repr__(self):
+        return 'EasyGUI PopUp Window'
+
+
 
 class Section(tk.Frame, GridMaster, SectionMaster):
     '''
@@ -328,13 +494,28 @@ class Section(tk.Frame, GridMaster, SectionMaster):
             self.tabs.style = self.style
             self.tabs.root = self.root
         self.equal_button_width = equal_button_width
-        self.widgets: dict = {}
         if title:  # title kwargs can be provided as True or a string
             if isinstance(title, str):  # if string, use title for label text
                 self.add_widget(type='label', text=title)
             elif title == True:  # if True, use the name as the label text
                 self.add_widget(type='label', text=name)
 
+    def __init_subclass__(cls, **kwargs):
+        '''
+        Wraps user subclass __init__ to implicitly handle the Section.__init__ call.
+        This avoids the need for subclass to use "super().__init__(*args, **kwargs)"
+        '''
+        old_init = cls.__init__  # reference to original subclass method so new_init isn't recursive
+        def new_init(self, *args, **kwargs):
+            Section.__init__(self, **kwargs)  # in place of super().__init__() in subclass __init__
+            try:
+                old_init(self, *args, **kwargs)
+            except TypeError:
+                # traceback.print_exc()
+                # print('\n\n* Are you subclassing Section or passing in kwargs to Section creation?\n* If so, remember to put a "**kwargs" in the __init__ function!\n')
+                old_init(self)
+        cls.__init__ = new_init  # overwrite subclass __init__ method
+
     @property
     def style(self):
         '''Goes upsteam to evenually reference EasyGUI.style'''
@@ -362,9 +543,10 @@ class Section(tk.Frame, GridMaster, SectionMaster):
     def match_child_button_widths(self):
         child_buttons = [child for child in self.widgets.values() if isinstance(child, widgets.Button)]
         if len(child_buttons) > 1:
-            max_width = int(round(max(child.width / 7.0 for child in child_buttons)))
+            max_width = int(round(max(child.width / 7.0 for child in child_buttons if not child.image)))
             for child in child_buttons:
-                child.config(width=max_width)
+                if not child.image:
+                    child.config(width=max_width)
 
     def position(self, force_row: bool=False) -> None:
         '''
@@ -392,43 +574,6 @@ class Section(tk.Frame, GridMaster, SectionMaster):
             print(f'\n--- GRID FAILED for Section: "{self.name}" ---\nTry ensuring "grid_area" arg is given for all Sections in a given parent.\nAdding to a new row instead.')
             self.parent.create(force_row=True)  # go back and fully recreate section forcing all children to be packed/in new rows
 
-    def add_widget(self, type='label', text='', widget_name=None, grid_area=None, **kwargs):
-        '''
-        Add a Widget object to this Section by calling the add_widget function in widgets.py
-        (Easier to keep the function there as it needs access to all the individual Widget classes.)
-        '''
-        return widgets.add_widget(self, type=type, text=text, widget_name=widget_name, grid_area=grid_area, **kwargs)
-
-    def delete_widget(self, widget_name) -> None:
-        '''
-        Fully delete a widget.
-        Pass without issue if the widget doesn't exist.
-        '''
-        try:
-            self.widgets[widget_name].destroy()
-            del self.widgets[widget_name]
-        except:
-            pass
-
-    def delete_all_widgets(self) -> None:
-        '''
-        Fully delete all child widgets of this section.
-        '''
-        for w_name in list(self.widgets.keys()):
-            self.delete_widget(w_name)
-
-
-    def _clear_and_recreate_plot(self, mpl_figure, widget_name, grid_area, kwargs):
-        old_widget = self.widgets[widget_name]  # grab reference to widget to be deleted so that its place in dict can be given to new widget
-
-        new_widget = self.add_widget(type='matplotlib', widget_name=widget_name, toolbar=old_widget.toolbar, grid_area=grid_area)
-        new_widget.position()
-        new_widget.draw_plot(mpl_figure=mpl_figure)
-        new_widget.position()  # have to reposition/create Widget
-
-        old_widget.destroy()  # destroy after new widget is positioned for slightly less flickering
-
-
     @property
     def width(self) -> float:
         '''
@@ -443,6 +588,9 @@ class Section(tk.Frame, GridMaster, SectionMaster):
         '''
         return float(sum(widget.height for widget in self.widgets.values()))
 
+    def __repr__(self) -> str:
+        return f'Section: "{self.name}"'
+
 
 
 
